@@ -31,6 +31,8 @@
 static const char * const paths_global[] = {
 	"global.uid",
 	"global.gid",
+	"global.username",
+	"global.groupname",
 	"global.count-threads",
 	"global.init-ssl",
 	"global.server-string",
@@ -45,6 +47,8 @@ static const char * const paths_global[] = {
 enum lejp_global_paths {
 	LEJPGP_UID,
 	LEJPGP_GID,
+	LEJPGP_USERNAME,
+	LEJPGP_GROUPNAME,
 	LEJPGP_COUNT_THREADS,
 	LWJPGP_INIT_SSL,
 	LEJPGP_SERVER_STRING,
@@ -63,6 +67,7 @@ static const char * const paths_vhosts[] = {
 	"vhosts[].port",
 	"vhosts[].interface",
 	"vhosts[].unix-socket",
+	"vhosts[].unix-socket-perms",
 	"vhosts[].sts",
 	"vhosts[].host-ssl-key",
 	"vhosts[].host-ssl-cert",
@@ -110,6 +115,16 @@ static const char * const paths_vhosts[] = {
 	"vhosts[].tls13-ciphers",
 	"vhosts[].client-tls13-ciphers",
 	"vhosts[].strict-host-check",
+
+	"vhosts[].listen-accept-role",
+	"vhosts[].listen-accept-protocol",
+	"vhosts[].apply-listen-accept", /* deprecates "onlyraw" */
+	"vhosts[].fallback-listen-accept",
+	"vhosts[].allow-non-tls",
+	"vhosts[].redirect-http",
+	"vhosts[].allow-http-on-https",
+
+	"vhosts[].disable-no-protocol-ws-upgrades",
 };
 
 enum lejp_vhost_paths {
@@ -119,6 +134,7 @@ enum lejp_vhost_paths {
 	LEJPVP_PORT,
 	LEJPVP_INTERFACE,
 	LEJPVP_UNIXSKT,
+	LEJPVP_UNIXSKT_PERMS,
 	LEJPVP_STS,
 	LEJPVP_HOST_SSL_KEY,
 	LEJPVP_HOST_SSL_CERT,
@@ -166,33 +182,16 @@ enum lejp_vhost_paths {
 	LEJPVP_TLS13_CIPHERS,
 	LEJPVP_CLIENT_TLS13_CIPHERS,
 	LEJPVP_FLAG_STRICT_HOST_CHECK,
-};
 
-static const char * const parser_errs[] = {
-	"",
-	"",
-	"No opening '{'",
-	"Expected closing '}'",
-	"Expected '\"'",
-	"String underrun",
-	"Illegal unescaped control char",
-	"Illegal escape format",
-	"Illegal hex number",
-	"Expected ':'",
-	"Illegal value start",
-	"Digit required after decimal point",
-	"Bad number format",
-	"Bad exponent format",
-	"Unknown token",
-	"Too many ']'",
-	"Mismatched ']'",
-	"Expected ']'",
-	"JSON nesting limit exceeded",
-	"Nesting tracking used up",
-	"Number too long",
-	"Comma or block end expected",
-	"Unknown",
-	"Parser callback errored (see earlier error)",
+	LEJPVP_LISTEN_ACCEPT_ROLE,
+	LEJPVP_LISTEN_ACCEPT_PROTOCOL,
+	LEJPVP_FLAG_APPLY_LISTEN_ACCEPT,
+	LEJPVP_FLAG_FALLBACK_LISTEN_ACCEPT,
+	LEJPVP_FLAG_ALLOW_NON_TLS,
+	LEJPVP_FLAG_REDIRECT_HTTP,
+	LEJPVP_FLAG_ALLOW_HTTP_ON_HTTPS,
+
+	LEJPVP_FLAG_DISABLE_NO_PROTOCOL_WS_UPGRADES,
 };
 
 #define MAX_PLUGIN_DIRS 10
@@ -201,6 +200,7 @@ struct jpargs {
 	struct lws_context_creation_info *info;
 	struct lws_context *context;
 	const struct lws_protocols *protocols;
+	const struct lws_protocols **pprotocols;
 	const struct lws_extension *extensions;
 	char *p, *end, valid;
 	struct lws_http_mount *head, *last;
@@ -212,6 +212,7 @@ struct jpargs {
 	const char **plugin_dirs;
 	int count_plugin_dirs;
 
+	unsigned int reject_ws_with_no_protocol:1;
 	unsigned int enable_client_ssl:1;
 	unsigned int fresh_mount:1;
 	unsigned int any_vhosts:1;
@@ -245,6 +246,15 @@ arg_to_bool(const char *s)
 	return 0;
 }
 
+static void
+set_reset_flag(unsigned int *p, const char *state, unsigned int flag)
+{
+	if (arg_to_bool(state))
+		*p |= flag;
+	else
+		*p &= ~(flag);
+}
+
 static signed char
 lejp_globals_cb(struct lejp_ctx *ctx, char reason)
 {
@@ -281,6 +291,12 @@ lejp_globals_cb(struct lejp_ctx *ctx, char reason)
 	case LEJPGP_GID:
 		a->info->gid = atoi(ctx->buf);
 		return 0;
+	case LEJPGP_USERNAME:
+		a->info->username = a->p;
+		break;
+	case LEJPGP_GROUPNAME:
+		a->info->groupname = a->p;
+		break;
 	case LEJPGP_COUNT_THREADS:
 		a->info->count_threads = atoi(ctx->buf);
 		return 0;
@@ -342,6 +358,7 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		const char *ss;
 
 		/* set the defaults for this vhost */
+		a->reject_ws_with_no_protocol = 0;
 		a->valid = 1;
 		a->head = NULL;
 		a->last = NULL;
@@ -369,6 +386,7 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		a->info->timeout_secs = i[3];
 
 		a->info->protocols = a->protocols;
+		a->info->pprotocols = a->pprotocols;
 		a->info->extensions = a->extensions;
 #if defined(LWS_WITH_TLS)
 		a->info->client_ssl_cipher_list = "ECDHE-ECDSA-AES256-GCM-SHA384:"
@@ -477,6 +495,12 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		}
 		a->any_vhosts = 1;
 
+		if (a->reject_ws_with_no_protocol) {
+			a->reject_ws_with_no_protocol = 0;
+
+			vhost->default_protocol_index = 255;
+		}
+
 #if defined(LWS_WITH_TLS)
 		if (a->enable_client_ssl) {
 			const char *cert_filepath =
@@ -571,6 +595,9 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		else
 			a->info->options &= ~(LWS_SERVER_OPTION_UNIX_SOCK);
 		return 0;
+	case LEJPVP_UNIXSKT_PERMS:
+		a->info->unix_socket_perms = a->p;
+		break;
 	case LEJPVP_STS:
 		if (arg_to_bool(ctx->buf))
 			a->info->options |= LWS_SERVER_OPTION_STS;
@@ -727,25 +754,19 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 #endif
 
 	case LEJPVP_NOIPV6:
-		if (arg_to_bool(ctx->buf))
-			a->info->options |= LWS_SERVER_OPTION_DISABLE_IPV6;
-		else
-			a->info->options &= ~(LWS_SERVER_OPTION_DISABLE_IPV6);
+		set_reset_flag(&a->info->options, ctx->buf,
+			       LWS_SERVER_OPTION_DISABLE_IPV6);
 		return 0;
 
 	case LEJPVP_FLAG_ONLYRAW:
-		if (arg_to_bool(ctx->buf))
-			a->info->options |= LWS_SERVER_OPTION_ONLY_RAW;
-		else
-			a->info->options &= ~(LWS_SERVER_OPTION_ONLY_RAW);
+		set_reset_flag(&a->info->options, ctx->buf,
+			    LWS_SERVER_OPTION_ADOPT_APPLY_LISTEN_ACCEPT_CONFIG);
 		return 0;
 
 	case LEJPVP_IPV6ONLY:
 		a->info->options |= LWS_SERVER_OPTION_IPV6_V6ONLY_MODIFY;
-		if (arg_to_bool(ctx->buf))
-			a->info->options |= LWS_SERVER_OPTION_IPV6_V6ONLY_VALUE;
-		else
-			a->info->options &= ~(LWS_SERVER_OPTION_IPV6_V6ONLY_VALUE);
+		set_reset_flag(&a->info->options, ctx->buf,
+			       LWS_SERVER_OPTION_IPV6_V6ONLY_VALUE);
 		return 0;
 
 	case LEJPVP_FLAG_CLIENT_CERT_REQUIRED:
@@ -755,20 +776,13 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		return 0;
 
 	case LEJPVP_IGNORE_MISSING_CERT:
-		if (arg_to_bool(ctx->buf))
-			a->info->options |= LWS_SERVER_OPTION_IGNORE_MISSING_CERT;
-		else
-			a->info->options &= ~(LWS_SERVER_OPTION_IGNORE_MISSING_CERT);
-
+		set_reset_flag(&a->info->options, ctx->buf,
+				LWS_SERVER_OPTION_IGNORE_MISSING_CERT);
 		return 0;
 
 	case LEJPVP_FLAG_STRICT_HOST_CHECK:
-		if (arg_to_bool(ctx->buf))
-			a->info->options |=
-				LWS_SERVER_OPTION_VHOST_UPG_STRICT_HOST_CHECK;
-		else
-			a->info->options &=
-				~(LWS_SERVER_OPTION_VHOST_UPG_STRICT_HOST_CHECK);
+		set_reset_flag(&a->info->options, ctx->buf,
+			LWS_SERVER_OPTION_VHOST_UPG_STRICT_HOST_CHECK);
 		return 0;
 
 	case LEJPVP_ERROR_DOCUMENT_404:
@@ -793,6 +807,40 @@ lejp_vhosts_cb(struct lejp_ctx *ctx, char reason)
 		a->info->alpn = a->p;
 		break;
 
+	case LEJPVP_LISTEN_ACCEPT_ROLE:
+		a->info->listen_accept_role = a->p;
+		break;
+	case LEJPVP_LISTEN_ACCEPT_PROTOCOL:
+		a->info->listen_accept_protocol = a->p;
+		break;
+
+	case LEJPVP_FLAG_APPLY_LISTEN_ACCEPT:
+		set_reset_flag(&a->info->options, ctx->buf,
+			LWS_SERVER_OPTION_ADOPT_APPLY_LISTEN_ACCEPT_CONFIG);
+		return 0;
+	case LEJPVP_FLAG_FALLBACK_LISTEN_ACCEPT:
+		lwsl_notice("vh %s: LEJPVP_FLAG_FALLBACK_LISTEN_ACCEPT: %s\n",
+			    a->info->vhost_name, ctx->buf);
+		set_reset_flag(&a->info->options, ctx->buf,
+		      LWS_SERVER_OPTION_FALLBACK_TO_APPLY_LISTEN_ACCEPT_CONFIG);
+		return 0;
+	case LEJPVP_FLAG_ALLOW_NON_TLS:
+		set_reset_flag(&a->info->options, ctx->buf,
+			       LWS_SERVER_OPTION_ALLOW_NON_SSL_ON_SSL_PORT);
+		return 0;
+	case LEJPVP_FLAG_REDIRECT_HTTP:
+		set_reset_flag(&a->info->options, ctx->buf,
+			       LWS_SERVER_OPTION_REDIRECT_HTTP_TO_HTTPS);
+		return 0;
+	case LEJPVP_FLAG_ALLOW_HTTP_ON_HTTPS:
+		set_reset_flag(&a->info->options, ctx->buf,
+			       LWS_SERVER_OPTION_ALLOW_HTTP_ON_HTTPS_LISTENER);
+		return 0;
+
+	case LEJPVP_FLAG_DISABLE_NO_PROTOCOL_WS_UPGRADES:
+		a->reject_ws_with_no_protocol = 1;
+		return 0;
+
 	default:
 		return 0;
 	}
@@ -829,7 +877,7 @@ lwsws_get_config(void *user, const char *f, const char * const *paths,
 {
 	unsigned char buf[128];
 	struct lejp_ctx ctx;
-	int n, m, fd;
+	int n, m = 0, fd;
 
 	fd = lws_open(f, O_RDONLY);
 	if (fd < 0) {
@@ -853,107 +901,41 @@ lwsws_get_config(void *user, const char *f, const char * const *paths,
 
 	if (m < 0) {
 		lwsl_err("%s(%u): parsing error %d: %s\n", f, n, m,
-			 parser_errs[-m]);
+			 lejp_error_to_string(m));
 		return 2;
 	}
 
 	return 0;
 }
 
-#if defined(LWS_WITH_LIBUV) && UV_VERSION_MAJOR > 0
+struct lws_dir_args {
+	void *user;
+	const char * const *paths;
+	int count_paths;
+	lejp_callback cb;
+};
 
 static int
-lwsws_get_config_d(void *user, const char *d, const char * const *paths,
-		   int count_paths, lejp_callback cb)
+lwsws_get_config_d_cb(const char *dirpath, void *user,
+		      struct lws_dir_entry *lde)
 {
-	uv_dirent_t dent;
-	uv_fs_t req;
+	struct lws_dir_args *da = (struct lws_dir_args *)user;
 	char path[256];
-	int ret = 0, ir;
-	uv_loop_t loop;
 
-	ir = uv_loop_init(&loop);
-	if (ir) {
-		lwsl_err("%s: loop init failed %d\n", __func__, ir);
-	}
-
-	if (!uv_fs_scandir(&loop, &req, d, 0, NULL)) {
-		lwsl_err("Scandir on %s failed\n", d);
-		return 2;
-	}
-
-	while (uv_fs_scandir_next(&req, &dent) != UV_EOF) {
-		lws_snprintf(path, sizeof(path) - 1, "%s/%s", d, dent.name);
-		ret = lwsws_get_config(user, path, paths, count_paths, cb);
-		if (ret)
-			goto bail;
-	}
-
-bail:
-	uv_fs_req_cleanup(&req);
-	while (uv_loop_close(&loop))
-		;
-
-	return ret;
-}
-
-#else
-
-#ifndef _WIN32
-static int filter(const struct dirent *ent)
-{
-	if (!strcmp(ent->d_name, ".") || !strcmp(ent->d_name, ".."))
+	if (lde->type != LDOT_FILE && lde->type != LDOT_UNKNOWN /* ZFS */)
 		return 0;
 
-	return 1;
+	lws_snprintf(path, sizeof(path) - 1, "%s/%s", dirpath, lde->name);
+
+	return lwsws_get_config(da->user, path, da->paths,
+				da->count_paths, da->cb);
 }
-#endif
-
-static int
-lwsws_get_config_d(void *user, const char *d, const char * const *paths,
-		   int count_paths, lejp_callback cb)
-{
-#if !defined(_WIN32) && !defined(LWS_WITH_ESP32)
-	struct dirent **namelist;
-	char path[256];
-	int n, i, ret = 0;
-
-	n = scandir(d, &namelist, filter, alphasort);
-	if (n < 0) {
-		lwsl_err("Scandir on %s failed\n", d);
-		return 1;
-	}
-
-	for (i = 0; i < n; i++) {
-		if (strchr(namelist[i]->d_name, '~'))
-			goto skip;
-		lws_snprintf(path, sizeof(path) - 1, "%s/%s", d,
-			 namelist[i]->d_name);
-		ret = lwsws_get_config(user, path, paths, count_paths, cb);
-		if (ret) {
-			while (i++ < n)
-				free(namelist[i]);
-			goto bail;
-		}
-skip:
-		free(namelist[i]);
-	}
-
-bail:
-	free(namelist);
-
-	return ret;
-#else
-	return 0;
-#endif
-}
-
-#endif
 
 int
 lwsws_get_config_globals(struct lws_context_creation_info *info, const char *d,
 			 char **cs, int *len)
 {
+	struct lws_dir_args da;
 	struct jpargs a;
 	const char * const *old = info->plugin_dirs;
 	char dd[128];
@@ -982,9 +964,13 @@ lwsws_get_config_globals(struct lws_context_creation_info *info, const char *d,
 			     LWS_ARRAY_SIZE(paths_global), lejp_globals_cb) > 1)
 		return 1;
 	lws_snprintf(dd, sizeof(dd) - 1, "%s/conf.d", d);
-	if (lwsws_get_config_d(&a, dd, paths_global,
-			       LWS_ARRAY_SIZE(paths_global),
-			       lejp_globals_cb) > 1)
+
+	da.user = &a;
+	da.paths = paths_global;
+	da.count_paths = LWS_ARRAY_SIZE(paths_global),
+	da.cb = lejp_globals_cb;
+
+	if (lws_dir(dd, &da, lwsws_get_config_d_cb) > 1)
 		return 1;
 
 	a.plugin_dirs[a.count_plugin_dirs] = NULL;
@@ -1000,6 +986,7 @@ lwsws_get_config_vhosts(struct lws_context *context,
 			struct lws_context_creation_info *info, const char *d,
 			char **cs, int *len)
 {
+	struct lws_dir_args da;
 	struct jpargs a;
 	char dd[128];
 
@@ -1011,6 +998,7 @@ lwsws_get_config_vhosts(struct lws_context *context,
 	a.valid = 0;
 	a.context = context;
 	a.protocols = info->protocols;
+	a.pprotocols = info->pprotocols;
 	a.extensions = info->extensions;
 
 	lws_snprintf(dd, sizeof(dd) - 1, "%s/conf", d);
@@ -1018,8 +1006,13 @@ lwsws_get_config_vhosts(struct lws_context *context,
 			     LWS_ARRAY_SIZE(paths_vhosts), lejp_vhosts_cb) > 1)
 		return 1;
 	lws_snprintf(dd, sizeof(dd) - 1, "%s/conf.d", d);
-	if (lwsws_get_config_d(&a, dd, paths_vhosts,
-			       LWS_ARRAY_SIZE(paths_vhosts), lejp_vhosts_cb) > 1)
+
+	da.user = &a;
+	da.paths = paths_vhosts;
+	da.count_paths = LWS_ARRAY_SIZE(paths_vhosts),
+	da.cb = lejp_vhosts_cb;
+
+	if (lws_dir(dd, &da, lwsws_get_config_d_cb) > 1)
 		return 1;
 
 	*cs = a.p;
